from scipy.stats import poisson, rankdata
import pandas as pd


def compute_fdr(df, total_chip_reads, total_input_reads, args):

    total_island_input_reads = df.Input.sum()

    # Hack needed in case we run on test data
    # TODO: why does SICER not need this? Different genome versions?
    # run with FDR=1 on original SICER and get this island:
    # chr7    61606400        61606799        3       2       0.167427550906  1.40719467956   0.1674275 50906
    # does it not show up with epic?
    if total_island_input_reads == 0:
        total_island_input_reads = 2

    scaling_factor = (total_chip_reads * 1.0) / total_input_reads

    zero_controls_multiplier = total_input_reads * 1.0 / args.effective_genome_length

    avg_0_denom = (df.End - df.Start + 1) * zero_controls_multiplier
    avg_0_denom[avg_0_denom > 0.25] = 0.25
    avg_0_denom = avg_0_denom * scaling_factor

    avg = df.Input * scaling_factor
    avg[df.Input == 0] = avg_0_denom[df.Input == 0]

    fold_change = df.ChIP / avg
    df.insert(len(df.columns), "Fold_change", fold_change)

    p_vals = pd.Series(poisson.sf(df.ChIP, avg), index=df.index)

    p_vals[df.ChIP <= avg] = 1
    df.insert(len(df.columns), "P", p_vals)

    ranked_p_values = rankdata(p_vals)
    fdr = p_vals * len(df) / ranked_p_values
    fdr[fdr > 1] = 1
    df.insert(len(df.columns), "FDR", fdr)

    df = df[df.FDR < args.false_discovery_rate_cutoff]

    return df



import logging
from sys import platform
from re import search, IGNORECASE
from io import BytesIO
from subprocess import check_output

import pandas as pd

from epic.config import logging_settings

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"


def find_readlength(args):
    """Estimate length of reads based on 10000 first."""

    bed_file = args.treatment[0]

    filereader = "cat "
    if bed_file.endswith(".gz") and search("linux", platform, IGNORECASE):
        filereader = "zcat "
    elif bed_file.endswith(".gz") and search("darwin", platform, IGNORECASE):
        filereader = "gzcat "
    elif bed_file.endswith(".bz2"):
        filereader = "bzgrep "
    elif bed_file.endswith(".bam"):
        filereader = "bamToBed -i "

    command = filereader + "{} | head -10000".format(bed_file)
    output = check_output(command, shell=True)

    df = pd.read_table(
        BytesIO(output),
        header=None,
        usecols=[1, 2],
        sep="\s+",
        names=["Start", "End"])

    readlengths = df.End - df.Start
    mean_readlength = readlengths.mean()
    median_readlength = readlengths.median()
    max_readlength = readlengths.max()
    min_readlength = readlengths.min()

    logging.info((
        "Used first 10000 reads of {} to estimate a median read length of {}\n"
        "Mean readlength: {}, max readlength: {}, min readlength: {}.").format(
            bed_file, median_readlength, mean_readlength, max_readlength,
            min_readlength))

    return median_readlength


def get_closest_readlength(estimated_readlength):
    """Find the predefined readlength closest to the estimated readlength.

    In the case of a tie, choose the shortest readlength."""

    readlengths = [36, 50, 75, 100]
    differences = [abs(r - estimated_readlength) for r in readlengths]
    min_difference = min(differences)
    index_of_min_difference = [i
                               for i, d in enumerate(differences)
                               if d == min_difference][0]

    return readlengths[index_of_min_difference]



import logging

from joblib import Parallel, delayed

try:
    from functools import lru_cache  # noqa
except ImportError:
    from functools32 import lru_cache  # noqa


def _merge_chip_and_input(chip_df, input_df):

    chip_df_nb_bins = len(chip_df)
    merged_df = chip_df.merge(input_df,
                              how="left",
                              on=["Chromosome", "Bin"],
                              suffixes=[" ChIP", " Input"])
    merged_df = merged_df[["Chromosome", "Bin", "Count ChIP", "Count Input"]]
    merged_df.columns = ["Chromosome", "Bin", "ChIP", "Input"]

    merged_df = merged_df.fillna(0)

    assert len(merged_df) == chip_df_nb_bins

    return merged_df


def merge_chip_and_input(chip_dfs, input_dfs, nb_cpu):

    logging.info("Merging ChIP and Input data.")
    merged_chromosome_dfs = Parallel(n_jobs=nb_cpu)(
        delayed(_merge_chip_and_input)(chip_df, input_df)
        for chip_df, input_df in zip(chip_dfs, input_dfs))
    return merged_chromosome_dfs


def get_total_number_of_reads(dfs):
    return sum([df.Count.sum() for df in dfs])


def merge_same_files(sample1_dfs, sample2_dfs, nb_cpu):

    logging.info("Merging same class data.")
    merged_chromosome_dfs = Parallel(n_jobs=nb_cpu)(
        delayed(_merge_same_files)(sample1_df, sample2_df)
        for sample1_df, sample2_df in zip(sample1_dfs, sample2_dfs))

    return merged_chromosome_dfs


def _merge_same_files(sample1_df, sample2_df):

    merged_df = sample1_df.merge(sample2_df,
                                 how="outer",
                                 on=["Chromosome", "Bin"])

    return merged_df.fillna(0)



""" Global settings for the logger module. """

import logging
import sys

logging.basicConfig(
    level=logging.DEBUG,
    format=
    '%(message)s (File: %(module)s, Log level: %(levelname)s, Time: %(asctime)s )',
    datefmt='%a, %d %b %Y %H:%M:%S',
    stream=sys.stderr)



WINDOW_P_VALUE = 0.20
BIN_SIZE = 0.001
E_VALUE = 1000
E_VALUE_THRESHOLD = E_VALUE * .0000001

from epic.statistics.compute_window_score import compute_window_score
from epic.statistics.compute_poisson import _poisson


def add_to_island_expectations_dict(
        average_window_readcount, current_max_scaled_score,
        island_eligibility_threshold, island_expectations, gap_contribution):
    """Can probably be heavily optimized.
    Time required to run can be seen from logging info."""

    scaled_score = current_max_scaled_score + E_VALUE
    for index in range(current_max_scaled_score + 1, scaled_score + 1):
        island_expectation = 0.0
        i = island_eligibility_threshold  #i is the number of tags in the added window

        current_island = int(round(index - compute_window_score(
            i, average_window_readcount) / BIN_SIZE))

        while (current_island >= 0):

            if current_island in island_expectations:
                island_expectation += _poisson(
                    i, average_window_readcount) * island_expectations[
                        current_island]
            i += 1
            current_island = int(round(index - compute_window_score(
                i, average_window_readcount) / BIN_SIZE))
        island_expectation *= gap_contribution
        if island_expectation:
            island_expectations[index] = island_expectation

    return island_expectations



from functools import partial
from io import BytesIO
from itertools import product
from logging import info
from subprocess import check_output, Popen, PIPE

import pandas as pd
from joblib import Parallel, delayed
from numpy import int32

from natsort import natsorted

from epic.config.genomes import create_genome_size_dict, get_genome_size_file
from epic.windows.count.merge_chromosome_dfs import merge_chromosome_dfs
from epic.windows.count.remove_out_of_bounds_bins import remove_out_of_bounds_bins


def count_reads_in_windows(bed_file, args):

    chromosome_size_dict = create_genome_size_dict(args.genome)
    chromosomes = natsorted(list(chromosome_size_dict.keys()))

    if not args.paired_end:
        parallel_count_reads = partial(_count_reads_in_windows, bed_file, args)
    else:
        parallel_count_reads = partial(_count_reads_in_windows_paired_end,
                                       bed_file, args)

    info("Binning chromosomes {}".format(", ".join([c.replace("chr", "")
                                                    for c in chromosomes])))
    chromosome_dfs = Parallel(n_jobs=args.number_cores)(
        delayed(parallel_count_reads)(chromosome_size_dict[chromosome],
                                      chromosome, strand)
        for chromosome, strand in product(chromosomes, ["+", "-"]))

    info("Merging the bins on both strands per chromosome.")
    both_chromosome_strand_dfs = [df_pair
                                  for df_pair in _pairwise(chromosome_dfs)]
    merged_chromosome_dfs = Parallel(
        n_jobs=args.number_cores)(delayed(merge_chromosome_dfs)(df_pair)
                                  for df_pair in both_chromosome_strand_dfs)

    return merged_chromosome_dfs


def _count_reads_in_windows(bed_file, args, chromosome_size, chromosome,
                            strand):

    halved_fragment_size = args.fragment_size // 2
    idx = 1 if strand == "+" else 2  # fragment start indices

    if not args.keep_duplicates:
        duplicate_handling = " uniq | "
    else:
        duplicate_handling = ""

    if bed_file.endswith(".gz"):
        grep = "zgrep "
    elif bed_file.endswith(".bz2"):
        grep = "bzgrep "
    elif bed_file.endswith(".bam"):
        grep = "bamToBed -i {} | grep".format(bed_file)
        bed_file = ""
    else:
        grep = "grep "

    command = (
        "{grep} -E '^{chromosome}\\b.*\\{strand}$' {bed_file} | "
        "cut -f 1-3,6 - | sort -k2,3n | {duplicate_handling} "
        "LC_ALL=C perl -a -ne '$F[{idx}]{strand}={halved_fragment_size};"  # shift fragments
        "$F[{idx}]=$F[{idx}]-($F[{idx}] % {window_size});"  # turn exact start into bin
        "print \"@F[0,{idx}]\n\"' | "
        "uniq -c | "
        "sed -e 's/^[ ]*//'").format(grep=grep,
                                     chromosome=chromosome,
                                     strand=strand,
                                     bed_file=bed_file,
                                     duplicate_handling=duplicate_handling,
                                     idx=idx,
                                     halved_fragment_size=halved_fragment_size,
                                     window_size=args.window_size)

    output = check_output(command, shell=True)

    out_table = pd.read_table(
        BytesIO(output),
        header=None,
        sep=" ",
        names=[bed_file, "Chromosome", "Bin"])

    out_table = remove_out_of_bounds_bins(out_table, chromosome_size)

    out_table[["Bin", bed_file]] = out_table[["Bin", bed_file]].astype(int32)

    return out_table


def count_reads_in_windows_paired_end(bed_file, args):

    chromosome_size_dict = create_genome_size_dict(args.genome)
    chromosomes = natsorted(list(chromosome_size_dict.keys()))

    parallel_count_reads = partial(_count_reads_in_windows_paired_end,
                                   bed_file, args)

    info("Binning chromosomes {}".format(", ".join([c.replace("chr", "")
                                                    for c in chromosomes])))
    chromosome_dfs = Parallel(n_jobs=args.number_cores)(
        delayed(parallel_count_reads)(chromosome_size_dict[chromosome],
                                      chromosome)
        for chromosome in chromosomes)
    print(chromosome_dfs[0])

    # info("Merging the bins on both strands per chromosome.")
    # both_chromosome_strand_dfs = [df_pair
    #                               for df_pair in _pairwise(chromosome_dfs)]
    # merged_chromosome_dfs = Parallel(
    #     n_jobs=args.number_cores)(delayed(merge_chromosome_dfs)(df_pair)
    #                               for df_pair in both_chromosome_strand_dfs)

    return chromosome_dfs


def _count_reads_in_windows_paired_end(bed_file, args, chromosome_size,
                                       chromosome):
    # takes too long time?

    if not args.keep_duplicates:
        duplicate_handling = " uniq | "
    else:
        duplicate_handling = ""

    command = """bamToBed -bedpe -i {bed_file} |
    grep -E "^{chromosome}\\b.*{chromosome}\\b.*" | # Both chromos must be equal; no chimeras (?)
    cut -f 1-6  | sort -k2,3n -k4,5n | {duplicate_handling} # get chr start end chr start end for PE; sort on location
    LC_ALL=C perl -a -ne 'use List::Util qw[min max]; $start = min($F[1], $F[2]); $end = max($F[4], $F[5]); $middle = $start + int(($end - $start)/2); $bin = $middle - $middle % 200; print "$F[0] $bin\\n"' | # Find bin of midpoint between start and ends
    uniq -c |
    sed -e 's/^[ ]*//'
    """.format(**locals())

    print(command)

    stdout, stderr = Popen(command,
                           stdout=PIPE,
                           stderr=PIPE,
                           shell=True).communicate()

    print(stderr.split()[:100], "stderr")
    print(stdout.decode().split()[:100], "stdout")

    out_table = pd.read_table(
        BytesIO(stdout),
        header=None,
        sep=" ",
        names=["Count", "Chromosome", "Bin"])

    out_table = remove_out_of_bounds_bins(out_table, chromosome_size)

    out_table[["Bin", "Count"]] = out_table[["Bin", "Count"]].astype(int32)

    return out_table


def _pairwise(iterable):
    col = iter(iterable)
    return zip(col, col)



from epic.utils.helper_functions import lru_cache
from numpy import log
from scipy.stats import poisson


@lru_cache()
def compute_window_score(i, poisson_parameter):

    # No enrichment; poisson param also average
    if i < poisson_parameter:
        return 0

    p_value = poisson.pmf(i, poisson_parameter)

    if p_value > 0:
        window_score = -log(p_value)
    else:
        # log of zero not defined
        window_score = 1000

    return window_score



"""Run whole epic pipeline."""

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"

from os.path import dirname
from sys import stdout
from itertools import chain
from collections import OrderedDict
from subprocess import call
import logging

import pandas as pd

from natsort import natsorted

from epic.windows.count.count_reads_in_windows import (
    count_reads_in_windows, count_reads_in_windows_paired_end)
from epic.config.genomes import create_genome_size_dict
from epic.statistics.compute_background_probabilites import compute_background_probabilities
from epic.statistics.count_to_pvalue import count_to_pvalue
from epic.statistics.fdr import compute_fdr
from epic.utils.helper_functions import merge_chip_and_input, get_total_number_of_reads, merge_same_files
from epic.windows.cluster.find_islands import find_islands


def run_epic(args):

    chip_windows = multiple_files_count_reads_in_windows(args.treatment, args)
    input_windows = multiple_files_count_reads_in_windows(args.control, args)

    chip_merged = _merge_files(chip_windows.values(), args.number_cores)
    input_merged = _merge_files(input_windows.values(), args.number_cores)

    if args.print_matrix:
        print_matrixes(chip_merged, input_merged, args.print_matrix)

    chip_merged = sum_columns(chip_merged)
    input_merged = sum_columns(input_merged)

    nb_chip_reads = get_total_number_of_reads(chip_merged)
    nb_input_reads = get_total_number_of_reads(input_merged)

    merged_dfs = merge_chip_and_input(chip_merged, input_merged,
                                      args.number_cores)

    score_threshold, island_enriched_threshold, average_window_readcount = \
        compute_background_probabilities(nb_chip_reads, args)

    dfs = count_to_pvalue(merged_dfs, island_enriched_threshold,
                          average_window_readcount, args.number_cores)

    dfs = find_islands(dfs, score_threshold, args)

    logging.info("Done finding islands.")
    logging.info("Concating dfs.")
    df = pd.concat([df for df in dfs if not df.empty])
    logging.info("Labeling island bins.")

    logging.info("Computing FDR.")
    df = compute_fdr(df, nb_chip_reads, nb_input_reads, args)

    # Just in case some ints got promoted to float somewhere
    df[["Start", "End", "ChIP", "Input"]] = df[["Start", "End", "ChIP", "Input"
                                                ]].astype(int)
    df.to_csv(stdout, index=False, sep=" ", na_rep="NA")

    return df.reset_index(
    )  # only returns a value to simplify integration tests


def sum_columns(dfs):

    new_dfs = []
    for df in dfs:
        s = df.set_index("Chromosome Bin".split())
        s = s.sum(axis=1)
        s.name = "Count"
        df = s.reset_index()
        new_dfs.append(df)

    return new_dfs


def print_matrixes(chip, input, outpath):

    logging.info("Concating chip matrix before printing.")

    dfc = pd.concat(chip, axis=0).reset_index(drop=True)
    dfc["Chromosome"] = dfc["Chromosome"].astype("category")
    dfc = dfc.set_index("Chromosome Bin".split(), append=True)

    logging.info("Concating input matrix before printing.")
    dfi = pd.concat(input, axis=0).reset_index(drop=True)
    dfi["Chromosome"] = dfi["Chromosome"].astype("category")
    dfi = dfi.set_index("Chromosome Bin".split(), append=True)

    dir = dirname(outpath)
    if dir:
        call("mkdir -p {}".format(dir), shell=True)

    chip_file = outpath + "_chip.csv"
    input_file = outpath + "_input.csv"

    logging.info("Writing chip matrix to file: " + chip_file)
    dfc.reset_index(level=0, drop=True).to_csv(chip_file, sep=" ", na_rep="NA")
    logging.info("Writing input matrix to file: " + input_file)
    dfi.reset_index(level=0,
                    drop=True).to_csv(input_file,
                                      sep=" ",
                                      na_rep="NA")


def get_island_bins(df, window_size, genome):
    """Finds the enriched bins in a df."""

    # need these chromos because the df might not have islands in all chromos
    chromosomes = natsorted(list(create_genome_size_dict(genome)))

    chromosome_island_bins = {}
    df_copy = df.reset_index(drop=False)
    for chromosome in chromosomes:
        cdf = df_copy.loc[df_copy.Chromosome == chromosome]
        if cdf.empty:
            chromosome_island_bins[chromosome] = set()
        else:
            island_starts_ends = zip(cdf.Start.values.tolist(),
                                     cdf.End.values.tolist())
            island_bins = chain(*[range(
                int(start), int(end), window_size)
                                  for start, end in island_starts_ends])
            chromosome_island_bins[chromosome] = set(island_bins)

    return chromosome_island_bins


def multiple_files_count_reads_in_windows(bed_files, args):
    """Use count_reads on multiple files and store result in dict.

    Untested since does the same thing as count reads."""

    bed_windows = OrderedDict()
    for bed_file in bed_files:
        logging.info("Binning " + bed_file)
        if args.paired_end:
            chromosome_dfs = count_reads_in_windows_paired_end(bed_file, args)
        else:
            chromosome_dfs = count_reads_in_windows(bed_file, args)
        bed_windows[bed_file] = chromosome_dfs

    return bed_windows


def _merge_files(windows, nb_cpu):
    """Merge lists of chromosome bin df chromosome-wise.

    windows is an OrderedDict where the keys are files, the values are lists of
    dfs, one per chromosome.

    Returns a list of dataframes, one per chromosome, with the collective count
    per bin for all files.

    TODO: is it faster to merge all in one command?
    """

    windows = iter(windows)  # can iterate over because it is an ordereddict
    merged = next(windows)

    # if there is only one file, the merging is skipped since the windows is used up
    for chromosome_dfs in windows:
        merged = merge_same_files(merged, chromosome_dfs, nb_cpu)

    return merged
